#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Institution User views module """
from django.db.models import Q
from django_filters import rest_framework as filters_dj
from django_filters.rest_framework import DjangoFilterBackend
from django.shortcuts import get_object_or_404
from django.utils import timezone
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import PermissionDenied
from rest_framework.views import Response
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import InstitutionUserSerializer
from tesla_ce.apps.api.v2.serializers import InstitutionCourseActivityExtendedSerializer
from tesla_ce.models import InstitutionUser
from tesla_ce.models import Activity
from tesla_ce.models.user import get_institution_user
from tesla_ce.models.user import is_global_admin


def roles_values():
    """ Get available roles for filter """
    return (
        ('ADMIN', 'ADMIN'),
        ('SEND', 'SEND'),
        ('LEGAL', 'LEGAL'),
        ('DATA', 'DATA'),
        ('INSTRUCTOR', 'INSTRUCTOR'),
        ('LEARNER', 'LEARNER'),
    )


class InstitutionUserFilter(filters_dj.FilterSet):
    """ Filter implementation for Institution Users """

    roles = filters_dj.MultipleChoiceFilter(field_name="roles",
                                            label="roles",
                                            choices=roles_values,
                                            method='filter_roles')

    username = filters_dj.CharFilter(lookup_expr='icontains')
    first_name = filters_dj.CharFilter(lookup_expr='icontains')
    last_name = filters_dj.CharFilter(lookup_expr='icontains')
    email = filters_dj.CharFilter(lookup_expr='icontains')
    uid = filters_dj.CharFilter(lookup_expr='icontains')

    def filter_roles(self, queryset, name, value):

        # No filter provided
        if value is None or len(value) == 0:
            return queryset

        # Filter by provided roles
        filter_q = Q()
        if 'ADMIN' in value:
            filter_q = filter_q | Q(inst_admin=True)
        if 'SEND' in value:
            filter_q = filter_q | Q(send_admin=True)
        if 'LEGAL' in value:
            filter_q = filter_q | Q(legal_admin=True)
        if 'DATA' in value:
            filter_q = filter_q | Q(data_admin=True)
        if 'LEARNER' in value:
            filter_q = filter_q | Q(learner__isnull=False)
        if 'INSTRUCTOR' in value:
            filter_q = filter_q | Q(instructor__isnull=False)
        return queryset.filter(filter_q)

    class Meta:
        model = InstitutionUser
        fields = ['username', 'first_name', 'last_name', 'email', 'roles', 'uid']


# pylint: disable=too-many-ancestors
class InstitutionUserViewSet(NestedViewSetMixin, viewsets.ModelViewSet):
    """
    API endpoint that allow institution users to be managed.
    """
    model = InstitutionUser
    serializer_class = InstitutionUserSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    filterset_class = InstitutionUserFilter
    search_fields = ['username', 'first_name', 'last_name', 'email', 'uid']
    permission_classes = [
        permissions.GlobalAdminReadOnlyPermission |
        permissions.InstitutionAdminPermission |
        permissions.InstitutionLegalAdminReadOnlyPermission |
        permissions.InstitutionDataAdminReadOnlyPermission |
        permissions.InstitutionMemberReadOnlyPermission
    ]

    def get_queryset(self):
        queryset = self.filter_queryset_by_parents_lookups(InstitutionUser.objects)
        if not is_global_admin(self.request.user):
            inst_user = get_institution_user(self.request.user)
            if not inst_user.inst_admin and not inst_user.legal_admin and not inst_user.data_admin:
                queryset = queryset.filter(id=inst_user.id)
        return queryset.all().order_by('id')

    @action(detail=True, methods=['GET'],
            serializer_class=InstitutionCourseActivityExtendedSerializer,
            permission_classes=[permissions.GlobalAdminReadOnlyPermission |
                                permissions.InstitutionAdminReadOnlyPermission |
                                permissions.InstitutionLegalAdminReadOnlyPermission |
                                permissions.InstitutionDataAdminReadOnlyPermission |
                                permissions.InstitutionMemberReadOnlyPermission]
            )
    def activities(self, request, *args, **kwargs):
        """
            Manage learner informed consent
        """
        # Ensure permissions
        if not is_global_admin(self.request.user):
            inst_user = get_institution_user(self.request.user)
            if not inst_user.inst_admin and not inst_user.legal_admin and not inst_user.data_admin \
                    and int(kwargs['pk']) != inst_user.id:
                raise PermissionDenied('You cannot access this information')

        user = get_object_or_404(InstitutionUser, id=kwargs['pk'])
        qs = Activity.objects.filter(vle__institution_id=kwargs['parent_lookup_institution_id'], enabled=True).filter(
            Q(course__instructors__id=user.id) | Q(course__learners__id=user.id)
        ).distinct().filter(
            Q(start__isnull=True) | Q(start__lte=timezone.now()),
            Q(end__isnull=True) | Q(end__gte=timezone.now())
        )
        page = self.paginate_queryset(qs)
        if page is not None:
            data = self.serializer_class(page, many=True, context={'request': self.request}).data
            return self.get_paginated_response(data)

        data = self.serializer_class(qs, many=True, context={'request': self.request}).data
        return Response(self.get_paginated_response(data))




#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA modules definition module"""


def get_modules():
    """
        Get the description of all modules

        :return: Dictionary with all modules and properties
        :rtype: dict
    """
    return {
        'api': get_api_properties(),
        'lapi': get_lapi_properties(),
        'worker': get_worker_properties(),
        'beat': get_beat_properties(),
        'worker-all': get_worker_all_properties(),
        'worker-enrolment': get_worker_enrol_properties(),
        'worker-enrolment-storage': get_worker_enrol_storage_properties(),
        'worker-enrolment-validation': get_worker_enrol_validation_properties(),
        'worker-verification': get_worker_verification_properties(),
        'worker-alerts': get_worker_alerts_properties(),
        'worker-reporting': get_worker_reporting_properties(),
    }


def get_api_properties():
    """
        Returns the definition of the API module

        :return: Definition of the API module
        :rtype: dict
    """
    return {
        'module': 'api',
        'description': 'TeSLA API',
        'apps': ['tesla_ce.apps.api'],
        'services': {
            'django': ['django/config'],
            'db': ['db/config', 'db/credentials/rw_access'],
            'vault': ['vault/jwt/learners/create', 'vault/jwt/learners/validate',
                      'vault/jwt/instructors/create', 'vault/jwt/instructors/validate',
                      'vault/jwt/module/all/validate', 'vault/jwt/module/all/create',
                      'vault/jwt/users/create', 'vault/jwt/users/validate',
                      'vault/vle/write', 'vault/policy/write', 'vault/provider/write',
                      ],
            'redis': ['redis/config'],
            'storage': ['storage/config'],
            'celery': ['celery/config', 'celery/queues', 'celery/credentials/api'],
        },
        'deployment': {
            'type': 'service',
            'public': True,
            'base_path': 'api',
        },
        'dependencies': [],
    }


def get_lapi_properties():
    """
        Returns the definition of the Learner API module

        :return: Definition of the Learner API module
        :rtype: dict
    """
    return {
        'module': 'lapi',
        'description': 'TeSLA Learner API',
        'apps': ['tesla_ce.apps.lapi'],
        'services': {
            'django': ['django/config'],
            'redis': ['redis/config'],
            'storage': ['storage/config'],
            'vault': ['vault/jwt/learners/validate', 'vault/jwt/users/validate'],
            'celery': ['celery/config', 'celery/queues', 'celery/credentials/lapi'],
        },
        'deployment': {
            'type': 'service',
            'public': True,
            'base_path': 'lapi',
        },
        'dependencies': [],
    }


def get_worker_properties():
    """
        Returns the definition of the Worker module

        :return: Definition of the worker module
        :rtype: dict
    """
    return {
        'module': 'worker',
        'description': 'TeSLA Worker',
        'apps': [],
        'services': {
            'django': ['django/config'],
            'db': ['db/config', 'db/credentials/rw_access'],
            'celery': ['celery/config', 'celery/queues', 'celery/credentials/worker'],
            'redis': ['redis/config'],
            'storage': ['storage/config'],
        },
        'deployment': {
            'type': 'worker',
            'public': False,
            'base_path': None,
            'queues': ['DEFAULT']
        },
        'dependencies': [],
    }


def get_worker_all_properties():
    """
        Returns the definition of the Worker process for all processing tasks

        :return: Definition of the worker module
        :rtype: dict
    """
    generic_worker = get_worker_properties()
    generic_worker['module'] = 'worker-all'
    generic_worker['description'] = 'TeSLA Worker performing all process tasks'
    generic_worker['deployment']['queues'] = ['__all__']
    return generic_worker


def get_worker_enrol_properties():
    """
        Returns the definition of the Worker process for generic enrolment tasks

        :return: Definition of the worker module
        :rtype: dict
    """
    generic_worker = get_worker_properties()
    generic_worker['module'] = 'worker-enrolment'
    generic_worker['description'] = 'TeSLA Worker performing generic enrolment tasks'
    generic_worker['deployment']['queues'] = ['ENROLMENT']
    return generic_worker


def get_worker_enrol_storage_properties():
    """
        Returns the definition of the Worker process devoted to store enrolment samples

        :return: Definition of the worker module
        :rtype: dict
    """
    generic_worker = get_worker_properties()
    generic_worker['module'] = 'worker-enrolment-storage'
    generic_worker['description'] = 'TeSLA Worker performing enrolment storage tasks'
    generic_worker['deployment']['queues'] = ['ENROLMENT_STORAGE']
    return generic_worker


def get_worker_enrol_validation_properties():
    """
        Returns the definition of the Worker process for enrolment samples validation tasks

        :return: Definition of the worker module
        :rtype: dict
    """
    generic_worker = get_worker_properties()
    generic_worker['module'] = 'worker-enrolment-validation'
    generic_worker['description'] = 'TeSLA Worker performing enrolment sample validation tasks'
    generic_worker['deployment']['queues'] = ['ENROLMENT_VALIDATION']
    return generic_worker


def get_worker_verification_properties():
    """
        Returns the definition of the Worker process for verification tasks

        :return: Definition of the worker module
        :rtype: dict
    """
    generic_worker = get_worker_properties()
    generic_worker['module'] = 'worker-verification'
    generic_worker['description'] = 'TeSLA Worker performing verification tasks'
    generic_worker['deployment']['queues'] = ['VERIFICATION']
    return generic_worker


def get_worker_reporting_properties():
    """
        Returns the definition of the Worker process for reporting tasks

        :return: Definition of the worker module
        :rtype: dict
    """
    generic_worker = get_worker_properties()
    generic_worker['module'] = 'worker-reporting'
    generic_worker['description'] = 'TeSLA Worker performing reporting tasks'
    generic_worker['deployment']['queues'] = ['REPORTING']
    return generic_worker


def get_worker_alerts_properties():
    """
        Returns the definition of the Worker process for alerts processing tasks

        :return: Definition of the worker module
        :rtype: dict
    """
    generic_worker = get_worker_properties()
    generic_worker['module'] = 'worker-alerts'
    generic_worker['description'] = 'TeSLA Worker performing alerts processing tasks'
    generic_worker['deployment']['queues'] = ['ALERTS']
    return generic_worker


def get_beat_properties():
    """
        Returns the definition of the Worker Beat module

        :return: Definition of the worker beat module
        :rtype: dict
    """
    return {
        'module': 'beat',
        'description': 'TeSLA Worker Beat',
        'apps': [],
        'services': {
            'django': ['django/config'],
            'db': ['db/config', 'db/credentials/rw_access'],
            'celery': ['celery/config', 'celery/queues', 'celery/credentials/beat'],
        },
        'deployment': {
            'type': 'beat',
            'public': False,
            'base_path': None,
        },
        'dependencies': [],
    }


def get_vle_properties():
    """
        Returns the definition of a VLE module

        :return: Definition of a VLE module
        :rtype: dict
    """
    return {
        'module': 'vle',
        'description': 'TeSLA VLE',
        'apps': [],
        'services': {
        },
        'deployment': {
            'type': None,
            'public': False,
            'base_path': None,
        },
        'dependencies': ['api', 'lapi'],
    }


def get_provider_properties():
    """
        Returns the definition of a Provider module

        :return: Definition of a Provider module
        :rtype: dict
    """
    return {
        'module': 'provider',
        'description': 'TeSLA Provider',
        'apps': [],
        'services': {
            'celery': ['celery/config', 'celery/credentials/provider'],
        },
        'deployment': {
            'type': None,
            'public': False,
            'base_path': None,
        },
        'dependencies': ['api'],
    }



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution course instructors management """
import logging

import pytest

import tests.utils


@pytest.mark.django_db
def test_api_institution_course_instructors(rest_api_client, user_global_admin, institution_course_test_case):
    # institution
    institution_id = institution_course_test_case['institution'].id
    # institution user
    institution_user = institution_course_test_case['user'].institutionuser
    # instructor
    instructor = institution_course_test_case['instructor']
    instructor_user = instructor.institutionuser
    instructor_user_id = instructor_user.id
    # learner
    learner = institution_course_test_case['learner']
    learner_user = learner.institutionuser
    learner_user_id = learner_user.id

    pytest.skip('TODO')

    # Get the list of courses for an instructor
    rest_api_client.force_authenticate(user=instructor_user)
    courses_url = '/api/v2/institution/{}/course/'.format(institution_id)
    str_module = 'List of Instructor\'s Courses (Instructor user privileges)'
    str_message = 'RESPONSE institution_id:{} instructor_user_id:{}'.format(institution_id, instructor_user_id)
    courses_list = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                   str_module, str_message, 200)
    n_courses = courses_list['count']
    assert n_courses == 1

    # Select first course
    course_id = courses_list['results'][0]['id']
    logging.info(course_id)
    # 666 TODO: add to code "id" vs "vle.id" vs "vle_course_id" vs "code"?!!

    # List learners from a course
    str_path_learner = '{}{}/learner/'.format(courses_url, course_id)
    str_module = 'List learners from a course (instructor user privileges)'
    str_message = 'RESPONSE institution_id:{} course_id:{} instructor_user_id:{}'.format(institution_id,
                                                                                         course_id,
                                                                                         instructor_user_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path_learner,
                                           str_module, str_message, 200)
    # number of learners
    n_learners = body['count']

    # TODO Add a Learner to a Course (from instructor privileges)
    # 666 STATUS 500 + KeyError('parent_lookup_institution_id')
    ''''''
    str_data = {'uid': institution_user.uid}
    new_learner_id = tests.utils.post_rest_api_client(rest_api_client, str_path_learner, str_data,
                                                      'Add new learner to a course', 'RESPONSE: ', 201)

    # Ensure number of learners has increased
    body = tests.utils.get_rest_api_client(rest_api_client, str_path_learner,
                                           'List learners from a course', 'RESPONSE:', 200)
    assert n_learners + 1 == body['count']
    ''''''

    # Check learner user can only list herself
    rest_api_client.force_authenticate(user=learner_user)
    # str_path = '/api/v2/institution/{}/course/{}/learner/'.format(institution_id, course_id)
    str_module = 'List learners from a course (learner user privileges)'
    str_message = 'RESPONSE institution_id:{} course_id:{} learner_user_id:{}'.format(institution_id,
                                                                                      course_id,
                                                                                      learner_user_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path_learner,
                                           str_module, str_message, 200)
    assert body['count'] == 1
    assert body['results'][0]['id'] == learner_user_id

    rest_api_client.force_authenticate(user=instructor_user)

    # TODO Remove a Learner from a Course but not from system (from instructor)
    '''
    # str_path = '/api/v2/institution/{}/course/{}/learner/{}/'.format(institution_id, course_id, new_learner_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path_learner,
                                       'Delete Learner from a Course', "RESPONSE: ", 204)
    assert body['count'] == n_learners
    
    # 666 TODO: check user still exist in the system?
    
    '''
    # List instructors from a course
    rest_api_client.force_authenticate(user=instructor_user)
    str_path_instructor = '{}{}/instructor/'.format(courses_url, course_id)
    str_module = 'List of Course\'s Instructors (Instructor user privileges)'
    str_message = 'RESPONSE institution_id:{} course_id:{} instructor_user_id:{}'.format(institution_id,
                                                                                         course_id,
                                                                                         instructor_user_id)
    instructors_list = tests.utils.get_rest_api_client(rest_api_client, str_path_instructor,
                                                       str_module, str_message, 200)
    n_instructors = instructors_list['count']
    assert n_instructors == 1
    assert instructors_list['results'][0]['id'] == instructor_user_id

    # TODO Add existing user as a instructor to a course (from instructor privileges)
    ''''''
    str_data = {'uid': institution_user.uid}
    new_instructor_id = tests.utils.post_rest_api_client(rest_api_client, str_path_instructor, str_data,
                                                         'Add new instructor to a course', 'RESPONSE: ', 201)

    # Ensure number of learners has increased
    body = tests.utils.get_rest_api_client(rest_api_client, str_path_instructor,
                                           'List instructors from a course', 'RESPONSE:', 200)
    assert n_instructors + 1 == body['count']
    ''''''
    # TODO Remove an instructor from course but not from system (from instructor privileges)
    '''
    # str_path = '/api/v2/institution/{}/course/{}/instructor/{}/'.format(institution_id, course_id, new_instructor_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path_instructor,
                                       'Delete Instructor from a Course', "RESPONSE: ", 204)
    assert body['count'] == n_instructors
    '''
    pytest.skip('TODO')



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" Queue Manager Module"""
from ..config import ConfigManager


class QueueManager:
    """ Manager class for Queues """

    #: Queue Client
    _client = None

    def __init__(self, config=None):
        """
            Default constructor
            :param config: Configuration manager instance
            :type config: ConfigManager
        """
        if config is None:
            config = ConfigManager()

        # Store the configuration object
        self._config = config

    @property
    def client(self):
        """
            Access to the Queue client
            :return: Queue client object
            :rtype: BaseQueueClient
        """
        if self._client is None:
            if self._config.get_celery_config():
                pass

        return self._client




#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""SENDCategory api serialize module."""
import json

from rest_framework import serializers

from tesla_ce.models import Instrument
from tesla_ce.models import SENDCategory
from .institution import InstitutionSerializer


def get_instruments():
    return [(inst.id, inst.name) for inst in Instrument.objects.all()]


class InstitutionSENDCategoryDataSerializer(serializers.Serializer):
    """SENDCategory serialize data module."""

    enabled_options = serializers.MultipleChoiceField(choices=['big_fonts', 'high_contrast', 'text_to_speech'],
                                                      required=False)
    disabled_instruments = serializers.MultipleChoiceField(choices=get_instruments(),
                                                           required=False)

    def to_internal_value(self, data):
        data_object = super().to_internal_value(data)
        data_object['enabled_options'] = list(data_object['enabled_options'])
        data_object['disabled_instruments'] = list(data_object['disabled_instruments'])

        return json.dumps(data_object)

    def to_representation(self, instance):
        if isinstance(instance, str):
            return json.loads(instance)
        return instance


class InstitutionSENDCategorySerializer(serializers.ModelSerializer):
    """SENDCategory serialize model module."""

    institution = InstitutionSerializer(read_only=True)
    data = InstitutionSENDCategoryDataSerializer(default={})
    institution_id = serializers.HiddenField(default=None, allow_null=True)

    class Meta:
        model = SENDCategory
        fields = "__all__"

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['institution_id'] = self.context['view'].kwargs['parent_lookup_institution_id']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for LAPI token based authentication """
import pytest


@pytest.mark.django_db
def test_lapi_authentication(rest_api_client, api_client, institution_course_test_case):

    # Get a launcher token for learner
    lapi_token = api_client.get_launcher_token('LAPI', institution_course_test_case['learner'].institutionuser)
    assert lapi_token is not None
    assert 'id' in lapi_token
    assert 'token' in lapi_token

    # Authenticate using the token
    resp_lapi_token = rest_api_client.post('/api/v2/auth/token', data=lapi_token)

    assert resp_lapi_token is not None
    assert resp_lapi_token.status_code == 200
    assert 'access_token' in resp_lapi_token.data
    assert 'refresh_token' in resp_lapi_token.data

    # Assign credentials
    token = resp_lapi_token.data['access_token']
    rest_api_client.credentials(HTTP_AUTHORIZATION='JWT {}'.format(token))

    # Use token to access LAPI
    lapi_call_resp = rest_api_client.get('/lapi/v1/profile/{}/{}/'.format(
        institution_course_test_case['learner'].institutionuser.institution_id,
        institution_course_test_case['learner'].institutionuser.learner.learner_id
    ))
    assert lapi_call_resp.status_code == 200



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA User module """
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .institution import Institution


def get_roles():
    return [
        "INSTRUCTOR",
        "LEARNER",
        "ADMIN",
        "SEND",
        "LEGAL",
        "DATA",
        "GLOBAL_ADMIN"
    ]


def get_institution_user(user):
    """
        Return a InstitutionUser object if user belongs to an institution or None otherwise
        :param user: User object
        :return: The InstitutionUser object or None
    """
    if isinstance(user, InstitutionUser):
        return user

    if isinstance(user, User):
        try:
            return user.institutionuser
        except User.institutionuser.RelatedObjectDoesNotExist:
            # If user has no institution this will fail
            return None
    return None


def is_global_admin(user):
    """
        Check if provided user is a Global Admin
        :param user: User object
        :return: True if it is a global admin or False otherwise
    """
    if isinstance(user, InstitutionUser):
        return user.user_ptr.is_staff

    if isinstance(user, User):
        return user.is_staff

    return False


def get_institution_roles(user):
    """
        Get the list of institution related roles for a user
        :param user: User object
        :return: list of roles
    """
    instance = get_institution_user(user)
    roles = []
    if instance is not None:
        if hasattr(instance, "instructor"):
            roles.append("INSTRUCTOR")
        if hasattr(instance, "learner"):
            roles.append("LEARNER")
        if instance.inst_admin:
            roles.append("ADMIN")
        if instance.send_admin:
            roles.append("SEND")
        if instance.legal_admin:
            roles.append("LEGAL")
        if instance.data_admin:
            roles.append("DATA")

    return roles


class BaseUser(BaseModel, User):
    """
        Base user for TeSLA CE.
    """
    locale = models.CharField(max_length=10, null=True, blank=False, default=None,
                              help_text=_('Default locale for this user'))
    created_at = models.DateTimeField(auto_now_add=True, help_text=_('Date when user was created'))
    updated_at = models.DateTimeField(auto_now=True, help_text=_('Last user modification'))

    def get_username(self):
        """
            Get the username of the user, corresponding to the email

            :return: User's username
            :rtype: str
        """
        return self.email

    def is_allowed(self, request):
        """
            Check if a request is allowed for current user, depending on the allowed scopes

            :return: Boolean value indicating if request is allowed for this user or not
        """
        for scope in self.allowed_scopes:
            if request.path.startswith(scope):
                return True
        return False

    @property
    def allowed_scopes(self):
        """
            Scopes the user is allowed to access

            :return: List of allowed scopes
            :rtype: list
        """
        return []

    class Meta:
        abstract = True

    def __repr__(self):
        return "<BaseUser(id='%r', email='%r', is_active='%r')>" % (
            self.id, self.email, self.is_active
        )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)


class UnauthenticatedUser(AnonymousUser):
    """
        Default user for unauthenticated requests
    """
    is_active = False

    def is_allowed(self, request):
        """
            Check if a request is allowed for current user, depending on the allowed scopes
            :return: Boolean value indicating if the user is authenticated or not
        """
        return False

    def __repr__(self):
        return "<UnauthenticatedUser>"


class AuthenticatedModule(AnonymousUser):
    """
        Default user for authenticated requests from modules
    """
    is_active = True
    module = None
    type = None

    def is_allowed(self, request):
        """
            Check if a request is allowed for current user, depending on the allowed scopes
            :return: Boolean value indicating if the user is authenticated or not
        """
        return False

    @property
    def is_authenticated(self):
        return True

    @property
    def is_anonymous(self):
        return False

    def __repr__(self):
        return "<AuthenticatedModule(type='%r', module='%r')>" % (
            self.type, self.module
        )


class InstitutionUser(BaseUser):
    """
        User belonging to an institution
    """
    uid = models.CharField(max_length=255, null=False, blank=False,
                           help_text=_('Unique User Identifier for the institution'))
    institution = models.ForeignKey(Institution, null=False, on_delete=models.CASCADE,
                                    help_text=_('Institution of the user'))

    login_allowed = models.BooleanField(null=False, default=False,
                                        help_text=_('Whether this user can login with user/password'))

    inst_admin = models.BooleanField(null=False, default=False,
                                     help_text=_('Whether this user is administrator of the institution'))

    legal_admin = models.BooleanField(null=False, default=False,
                                      help_text=_('Whether this user can manage legal data of the institution'))

    send_admin = models.BooleanField(null=False, default=False,
                                     help_text=_('Whether this user can manage SEND data of the institution'))

    data_admin = models.BooleanField(null=False, default=False,
                                     help_text=_('Whether this user can manage the data of the institution'))

    class Meta:
        unique_together = (('institution', 'uid'),)


class AuthenticatedUser(AnonymousUser):
    """
        Default user for authenticated requests from JWT without DB access
    """
    is_active = True
    type = None
    allowed_scopes = None
    sub = None
    filters = None

    def is_allowed(self, request):
        """
            Check if a request is allowed for current user, depending on the allowed scopes
            :return: Boolean value indicating if the user is authenticated or not
        """
        return False

    def is_authenticated(self):
        return True

    def __repr__(self):
        return "<AuthenticatedUser(type='%r', sub='%r')>" % (
            self.type, self.sub
        )



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" Swarm deployment scripts management tests """
import os
import tempfile
import mock

from io import StringIO

from django.core.management import call_command
import pytest


def test_swarm_services_deployment(tesla_ce_system):

    pytest.skip('TODO')

    out = StringIO()
    err = StringIO()

    # If local configuration file is not available, generate configuration
    if not os.path.exists('tesla-ce.cfg'):
        assert tesla_ce_system.config.get('VAULT_TOKEN') is not None
        # Write the configuration file to disk
        with open('tesla-ce.cfg', 'w') as out_fh:
            tesla_ce_system.config.config.write(out_fh)
    else:
        from tesla_ce.lib.config import ConfigManager
        conf = ConfigManager(load_config=False)
        conf.load_file('tesla-ce.cfg')
        assert conf.config.get('VAULT_TOKEN') is not None

    # Enable with services flag
    deploy_services = tesla_ce_system.config.config.get('DEPLOYMENT_SERVICES')
    tesla_ce_system.config.config.set('DEPLOYMENT_SERVICES', True)
    with tempfile.TemporaryDirectory() as tmp_dir:
        assert tmp_dir is not None
        with mock.patch('tesla_ce.management.base.TeslaCommand.get_client', return_value=tesla_ce_system):
            call_command(
                'deploy_services',
                stdout=out,
                stderr=err,
                out=tmp_dir,
                mode='swarm'
            )

        gen_files = os.listdir(tmp_dir)

        assert 'secrets' in gen_files
        assert os.path.isdir(os.path.join(tmp_dir, 'secrets'))
        assert len(os.listdir(os.path.join(tmp_dir, 'secrets'))) > 0

        assert 'config' in gen_files
        assert os.path.isdir(os.path.join(tmp_dir, 'secrets'))
        assert len(os.listdir(os.path.join(tmp_dir, 'secrets'))) > 0

    assert 'tesla_lb.yml' in gen_files
    assert 'tesla_services.yml' in gen_files
    tesla_ce_system.config.config.set('DEPLOYMENT_SERVICES', deploy_services)


def test_swarm_core_deployment(tesla_ce_system):

    out = StringIO()
    err = StringIO()

    # If local configuration file is not available, generate configuration
    if not os.path.exists('tesla-ce.cfg'):
        assert tesla_ce_system.config.get('VAULT_TOKEN') is not None
        # Write the configuration file to disk
        with open('tesla-ce.cfg', 'w') as out_fh:
            tesla_ce_system.config.config.write(out_fh)
    else:
        from tesla_ce.lib.config import ConfigManager
        conf = ConfigManager(load_config=False)
        conf.load_file('tesla-ce.cfg')
        assert conf.config.get('VAULT_TOKEN') is not None

    with mock.patch('tesla_ce.management.base.TeslaCommand.get_client', return_value=tesla_ce_system):
        with tempfile.TemporaryDirectory() as tmp_dir:
            call_command(
                'deploy_core',
                stdout=out,
                stderr=err,
                out=tmp_dir,
                mode='swarm'
            )

            gen_files = os.listdir(tmp_dir)

            assert 'secrets' in gen_files
            assert os.path.isdir(os.path.join(tmp_dir, 'secrets'))
            assert len(os.listdir(os.path.join(tmp_dir, 'secrets'))) > 0

        assert 'tesla_lb.yml' in gen_files
        assert 'tesla_core.yml' in gen_files



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango command to register a new Provider to TeSLA system """
import json
import os

from django.conf import settings
from django.core.management.base import CommandError

from ..base import TeslaDeployCommand


class Command(TeslaDeployCommand):
    """ Command to register Providers and optionally generate their deployment scripts """
    help = 'Register a new Provider to TeSLA CE system'
    requires_system_checks = '__all__'

    def add_arguments(self, parser):
        """
                    Define custom arguments for this command

                    :param parser: Input command parser instance
                """
        # Set default arguments
        super().add_arguments(parser)

        parser.add_argument(
            '--force-update',
            action='store_true',
            help='If the Provider already exist, update the information and generate new credentials',
        )
        parser.add_argument(
            '--provider_file',
            help='Provider description file',
        )
        parser.add_argument(
            '--deploy',
            action='store_true',
            help='Create deployment script',
        )

    def custom_handle(self):
        """
            Custom actions for this command
        """

        # Check if a provider description file is provided
        provider_file = self._options['provider_file']
        if provider_file is not None:
            if os.path.exists(provider_file):
                try:
                    with open(provider_file, 'r') as fd_provider:
                        provider_desc = json.load(fd_provider)
                        self.stdout.write('Reading provider description from {}: {}'.format(provider_file,
                                                                                            self.style.SUCCESS('[OK]')))
                except Exception as exc:
                    self.stdout.write('Reading provider description from {}: {}: {}'.format(provider_file,
                                                                                            self.style.ERROR('[ERROR]'),
                                                                                            exc))
            else:
                self.stdout.write('Reading configuration from {}: {}'.format(provider_file,
                                                                             self.style.ERROR('[ERROR]')))
                raise CommandError('Configuration file not found')
        else:
            # List registered providers
            providers = self.client.get_registered_providers()
            if len(providers) == 0:
                raise CommandError('Cannot find public registered providers on repository')
            provider_desc = None
            while provider_desc is None:
                self.stdout.write('Enter the acronym of the provider to be registered: ')
                available_options = []
                descriptions = {}
                for provider in providers:
                    self.stdout.write('  [{}] {}'.format(provider['acronym'], provider['name']))
                    available_options.append(provider['acronym'])
                    descriptions[provider['acronym']] = provider
                selected_provider = input('Enter the acronym: ')
                if selected_provider not in available_options:
                    self.stdout.write('{}'.format(self.style.ERROR('Invalid option')))
                else:
                    provider_desc = descriptions[selected_provider]

        # Get arguments
        force_update = self._options['force_update']

        # Register the new VLE
        provider_info = self.client.register_provider(provider_desc, force_update)
        self.stdout.write('{}. Use this configuration: {}'.format(
            self.style.SUCCESS('Provider registered'),
            json.dumps(provider_info, indent=4, sort_keys=True))
        )

        if 'url' in provider_desc:
            self.stdout.write('\nCheck deployment instructions on {}\n'.format(provider_desc['url']))

        if self._options['deploy']:
            # If provider have credentials, request values
            cred_list = None
            if 'credentials' in provider_desc and len(provider_desc['credentials']) > 0:
                cred_list = []
                self.stdout.write('This provider requires additional configuration. Enter value for:')
                for credential in provider_desc['credentials']:
                    req_text = ''
                    if 'required' in provider_desc and credential in provider_desc['required']:
                        req_text = '[REQUIRED] '
                    cred_value = None
                    while cred_value is None:
                        cred_value = input('{} {}: '.format(req_text, credential.upper()))
                        if len(cred_value.strip()) == 0 and len(req_text) > 0:
                            cred_value = None
                            self.stdout.write('   {} This parameter is required.Use this configuration'.format(
                                self.style.ERROR('ERROR')
                            ))
                    if len(cred_value.strip()) > 0:
                        cred_list.append({
                            'name': credential,
                            'value': cred_value
                        })

            # Export the deployment scripts
            self.client.export_provider_scripts(
                provider_desc['acronym'],
                credentials=cred_list,
                output=self._options['out'],
                mode=self._options['mode'])
            self.stdout.write(self.style.SUCCESS('Deployment scripts written at {}'.format(self._options['out'])))



# Generated by Django 3.0.8 on 2020-07-30 09:50

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0008_auto_20200721_1022'),
    ]

    operations = [
        migrations.CreateModel(
            name='Monitor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(0, 'Unknown'), (1, 'Ready'), (2, 'Error'), (3, 'Warning')], default=0, help_text='Status of the resource')),
                ('type', models.IntegerField(choices=[(0, 'Custom'), (1, 'Queue'), (2, 'Service'), (3, 'Provider'), (4, 'Instrument')], default=0, help_text='Type of the resource')),
                ('key', models.CharField(default=None, help_text='key of the resource', max_length=15, null=True)),
                ('info', models.TextField(help_text='Resource information.')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
